from repacolors import ColorScale
from .colorbrewer import PALETTES as CBPALETTES


PALETTES = {
    "ryb": ["#fe2713", "#fd5307", "#fb9900", "#fabc00", "#fefe34", "#d1e92c", "#66b032", "#0492ce", "#0347fe", "#3e01a4", "#8600af", "#a7194b"],
    "rybw3": ["#FE2712", "#FC600A", "#FB9902", "#FCCC1A", "#FEFE33", "#B2D732", "#66B032", "#347C98", "#0247FE", "#4424D6", "#8601AF", "#C21460"],
    **CBPALETTES
}


def get_palette(name: str):
    if name.lower() not in PALETTES:
        raise KeyError(f"'{name}' palette not found")

    return PALETTES[name.lower()]


def get_scale(name: str, *args, **kwargs) -> ColorScale:
    kwargs["name"] = name
    return ColorScale(get_palette(name), *args, **kwargs)


def demo(width: int = 80):
    for name, colors in PALETTES.items():
        s = ColorScale(colors)
        print(f"{name:12s}", end="")
        s.print(width=width, height=2, border=0)



from typing import Iterable, Any, Union, List
from itertools import zip_longest
import shutil


def _linepairs(image: Iterable[Iterable[Any]]):
    i = iter(image)
    return zip_longest(i, i)


class TerminalColor():
    """Abstract class for terminal pixel
    """

    def __init__(self, color):
        self.color = color

    @property
    def termbg(self):
        return f"\x1b[48;5;{self.color}m"

    @property
    def termfg(self):
        return f"\x1b[38;5;{self.color}m"

    termreset = "\x1b[0m"


def termsize():
    return shutil.get_terminal_size((80, 20))


def draw(image: Iterable[Iterable[TerminalColor]]) -> str:
    """Draw image (with ANSI escape sequences)

    image - TerminalPixel[][], should be a rectangle
    """
    output = [""]

    linepairs = _linepairs(image)

    for l1, l2 in linepairs:
        if l2 is None:
            l2 = []
        for c1, c2 in zip_longest(l1, l2):
            if c1 and c2:
                output.append(f"{c1.termfg}{c2.termbg}▀")
            elif c1:
                output.append(f"{c1.termfg}▀")
            elif c2:
                output.append(f"{c2.termbg}▀")
        output.append(TerminalColor.termreset)
        output.append("\n")

    return "".join(output)



"""Conversion between different color spaces

http://easyrgb.com/en/math.php
"""

import colorsys
import math
from typing import Tuple, NamedTuple
from .types import *
from .distance import distance

ANSI16 = [
    (0, 0, 0),
    (0.5, 0, 0),
    (0, 0.5, 0),
    (0.5, 0.5, 0),
    (0, 0, 0.5),
    (0.5, 0, 0.5),
    (0, 0.5, 0.5),
    (0.75, 0.75, 0.75),
    (0.5, 0.5, 0.5),
    (1, 0, 0),
    (0, 1, 0),
    (1, 1, 0),
    (0, 0, 1),
    (1, 0, 1),
    (0, 1, 1),
    (1, 1, 1),
]

# whitepoints
D50 = (96.422, 100.000, 82.521)
D55 = (95.682, 100.000, 92.149)
D65 = (95.047, 100.000, 108.883)
D75 = (94.972, 100.000, 122.638)


def ansi2rgb(ansicolor: int) -> RGBTuple:
    if 0 <= ansicolor < 16:
        # first 16 colors - special case
        return RGBTuple(*ANSI16[ansicolor])
    elif 16 <= ansicolor < 232:
        # colors
        i = ansicolor - 16
        ar = i // 36
        ag = (i - ar * 36) // 6
        ab = i - ar * 36 - ag * 6
        return RGBTuple(*((55 + 40 * x) / 255 if x > 0 else 0 for x in (ar, ag, ab)))
    elif 232 <= ansicolor < 256:
        # grayscale
        i = ansicolor - 231
        c = ((256 / 32 * i) + (i - 1) * 2) / 255
        return RGBTuple(c, c, c)

    # out of range
    return RGBTuple(0, 0, 0)


def rgb2ansi(color: CTuple) -> int:
    r, g, b = color
    coli = tuple(float(int(255 * x)) for x in (r, g, b))
    steps = [0] + [95 + 40 * x for x in range(5)]

    res = []
    for c in coli:
        for i in range(len(steps)):
            p, n = steps[i], steps[i + 1]
            if p <= c <= n:
                dp, dn = abs(p - c), abs(n - c)
                if dp < dn:
                    closest = i
                else:
                    closest = i + 1
                res.append(float(closest))
                break

    ansicol = 16 + sum(x[1] * 6 ** (2 - x[0]) for x in enumerate(res))

    # check grayscale
    if (
        abs(coli[0] - coli[1]) < 10
        and abs(coli[1] - coli[2]) < 10
        and abs(coli[0] - coli[2]) < 10
    ):
        avg = sum(coli) / 3
        for i in range(24):
            cc = (256 / 32 * (i + 1)) + i * 2
            if abs(cc - avg) <= 5:
                ansigray = 232 + i
                c1 = ansi2rgb(ansicol)
                c2 = ansi2rgb(ansigray)
                if distance(color, c1) > distance(color, c2):
                    ansicol = ansigray

    return int(ansicol)


def wavelength2rgb(wl: float, gamma: float = .8, darken: bool = True) -> RGBTuple:
    # TODO ir / uv fade out
    if wl < 380:
        return RGBTuple(0, 0, 0)
    elif wl < 440:
        c = 1 if not darken else (wl - 380) / 60
        s = (.3 + .7 * (wl - 380) / 40) * c
        return RGBTuple((s * (440 - wl) / 40) ** gamma, 0, s ** gamma)
    elif wl < 490:
        return RGBTuple(0, ((wl - 440) / 50) ** gamma, 1)
    elif wl < 510:
        return RGBTuple(0, 1, ((510 - wl) / 20) ** gamma)
    elif wl < 580:
        return RGBTuple(((wl - 510) / 70) ** gamma, 1, 0)
    elif wl < 645:
        return RGBTuple(1, ((645 - wl) / 65) ** gamma, 0)
    elif wl < 750:
        c = 1 if not darken else (750 - wl) / 105
        s = (.3 + .7 * (750 - wl) / 105) * c
        return RGBTuple(s ** gamma, 0, 0)

    # too big
    return RGBTuple(0, 0, 0)


def rgb2hex(color: CTuple, force_long: bool = False) -> str:
    r, g, b = tuple(max(0, min(1, c)) for c in color)
    hx = "".join(["%02x" % int((c + .0025) * 255) for c in (r, g, b)])

    if not force_long and hx[0::2] == hx[1::2]:
        hx = "".join(hx[0::2])

    return "#%s" % hx


def hex2rgb(hexcolor: str) -> RGBTuple:
    try:
        rgb = hexcolor[1:]

        if len(rgb) == 6 or len(rgb) == 8:
            r, g, b = rgb[0:2], rgb[2:4], rgb[4:6]
        elif len(rgb) == 3 or len(rgb) == 4:
            r, g, b = rgb[0] * 2, rgb[1] * 2, rgb[2] * 2
        else:
            raise ValueError()
    except Exception:
        raise ValueError("Invalid value %r provided for rgb color." % hexcolor)

    return RGBTuple(*(float(int(v, 16)) / 255 for v in (r, g, b)))


def rgb2yuv(color: RGBTuple) -> YUVTuple:
    r, g, b = color
    y = 0.299 * r + 0.587 * g + 0.114 * b
    u = - 0.168736 * r - 0.331364 * g + 0.5 * b
    v = 0.5 * r - 0.418688 * g - 0.081312 * b
    return YUVTuple(y, u, v)


def yuv2rgb(color: YUVTuple) -> RGBTuple:
    y, u, v = color
    r = y + 1.402 * v
    g = y - 0.34414 * u - 0.71414 * v
    b = y + 1.772 * u
    return RGBTuple(r, g, b)


def rgb2hsl(color: RGBTuple) -> HSLTuple:
    h, l, s = colorsys.rgb_to_hls(*color)
    return HSLTuple(h, s, l)


def hsl2rgb(color: HSLTuple) -> RGBTuple:
    h, s, l = color
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    return RGBTuple(r, g, b)


def hsv2hwb(color: HSVTuple) -> HWBTuple:
    return HWBTuple(color[0], (1 - color[1]) * color[2], 1 - color[2])


def hwb2hsv(color: HWBTuple) -> HSVTuple:
    return HSVTuple(color[0], 1 if color[2] == 1 else 1 - (color[1] / (1 - color[2])), 1 - color[2])


def rgb2xyz(color: RGBTuple) -> XYZTuple:
    vr, vg, vb = tuple(
        ((c + 0.055) / 1.055) ** 2.4 if c > 0.04045 else c / 12.92 for c in color
    )

    x = 0.4124564 * vr + 0.3575761 * vg + 0.1804375 * vb
    y = 0.2126729 * vr + 0.7151522 * vg + 0.0721750 * vb
    z = 0.0193339 * vr + 0.1191920 * vg + 0.9503041 * vb

    return XYZTuple(x, y, z)


def xyz2rgb(color: XYZTuple) -> RGBTuple:
    x, y, z = color

    vr = x * 3.2404542 + y * -1.5371385 + z * -0.4985314
    vg = x * -0.9692660 + y * 1.8760108 + z * 0.0415560
    vb = x * 0.0556434 + y * -0.2040259 + z * 1.0572252

    r, g, b = tuple(
        1.055 * c ** (1 / 2.4) - 0.055 if c > 0.0031308 else 12.92 * c
        for c in (vr, vg, vb)
    )

    return RGBTuple(r, g, b)


def xyz2lab(color: XYZTuple, whitepoint: CTuple = D65) -> LabTuple:
    x, y, z = tuple(100 * c[0] / c[1] for c in zip(color, whitepoint))

    vx, vy, vz = tuple(
        c ** (.33333333) if c > 0.008856452 else 7.787 * c + 0.137931034 for c in (x, y, z)
    )

    return LabTuple(116 * vy - 16, 500 * (vx - vy), 200 * (vy - vz))


def lab2xyz(color: LabTuple, whitepoint: CTuple = D65) -> XYZTuple:
    l, a, b = color
    vy = (l + 16) / 116
    vx = a / 500 + vy
    vz = vy - b / 200

    x, y, z = tuple(
        c ** 3 if c > 0.206896552 else (c - 0.137931034) / 7.787 for c in (vx, vy, vz)
    )

    return XYZTuple(x * whitepoint[0] / 100, y * whitepoint[1] / 100, z * whitepoint[2] / 100)


def lab2lch(color: LabTuple) -> LChTuple:
    l, a, b = color
    c = (a ** 2 + b ** 2) ** .5
    h = (math.atan2(b, a) / (math.pi * 2)) % 1  # convert radians to 0 - 1

    return LChTuple(l, c, h)


def lch2lab(color: LChTuple) -> LabTuple:
    l, c, h = color
    a = c * math.cos(h * 2 * math.pi)
    b = c * math.sin(h * 2 * math.pi)

    return LabTuple(l, a, b)


def rgb2lab(color: RGBTuple, whitepoint: CTuple = D65) -> LabTuple:
    return xyz2lab(rgb2xyz(color), whitepoint)


def lab2rgb(color: LabTuple, whitepoint: CTuple = D65) -> RGBTuple:
    return xyz2rgb(lab2xyz(color, whitepoint))


def rgb2lch(color: RGBTuple, whitepoint: CTuple = D65) -> LChTuple:
    return lab2lch(rgb2lab(color, whitepoint))


def lch2rgb(color: LChTuple, whitepoint: CTuple = D65) -> RGBTuple:
    return lab2rgb(lch2lab(color), whitepoint)


def rgb2cmyk(color: RGBTuple) -> CMYKTuple:
    vc, vm, vy = tuple(1 - v for v in color)
    k = 1.0
    for v in (vc, vm, vy):
        if v < k:
            k = v

    if k == 1.0:
        print("cica")
        c, m, y = 0.0, 0.0, 0.0
    else:
        c, m, y = tuple((v - k) / (1 - k) for v in (vc, vm, vy))

    return CMYKTuple(c, m, y, k)


def cmyk2rgb(color: CMYKTuple) -> RGBTuple:
    c, m, y, k = color
    return RGBTuple(*(1 - (v * (1 - k) + k) for v in (c, m, y)))


# add these as well since we have them for "free"
def rgb2yiq(color: RGBTuple) -> CTuple:
    return colorsys.rgb_to_yiq(*color)


def yiq2rgb(color: CTuple) -> RGBTuple:
    return RGBTuple(*colorsys.yiq_to_rgb(*color))


def rgb2hsv(color: RGBTuple) -> HSVTuple:
    return HSVTuple(*colorsys.rgb_to_hsv(*color))


def hsv2rgb(color: HSVTuple) -> RGBTuple:
    return RGBTuple(*colorsys.hsv_to_rgb(*color))


def rgb2hwb(color: RGBTuple) -> HWBTuple:
    return hsv2hwb(rgb2hsv(color))


def hwb2rgb(color: HWBTuple) -> RGBTuple:
    return hsv2rgb(hwb2hsv(color))



import os
import xcffib
import xcffib.xproto
from . import Theme


def get(DISPLAY=None, defaults=None):
    """
    Get the X resources
    """

    if DISPLAY is None:
        DISPLAY = os.environ.get("DISPLAY")

    if defaults is None:
        resources = {}
    else:
        resources = defaults.copy()

    try:
        conn = xcffib.connect(display=DISPLAY)
    except xcffib.ConnectionException as err:
        print(err)
        return resources

    root = conn.get_setup().roots[0].root
    atom = conn.core.InternAtom(False, 16, 'RESOURCE_MANAGER').reply().atom

    reply = conn.core.GetProperty(
        False, root, atom, xcffib.xproto.Atom.STRING,
        0, (2 ** 32) - 1).reply()
    conn.disconnect()

    resource_string = reply.value.buf().decode("utf-8")
    resource_list = filter(None, resource_string.split("\n"))

    for resource in resource_list:
        key, value = resource.split(":\t")
        resources[key.strip("*.")] = value

    return resources


_COLOR_PROPS = [f"color{i}" for i in range(16)] + ["background", "foreground"]


def get_colors(DISPLAY=None):
    resources = get(DISPLAY)

    return [resources.get(key, "#000000") for key in _COLOR_PROPS]


def get_theme(DISPLAY=None):
    colors = get_colors(DISPLAY)
    xtheme = Theme(colors, _COLOR_PROPS.copy())

    return xtheme



from typing import Iterator, List, Union, Callable
import operator
from . import colors
from .types import *

def equal_hex(c1: "colors.Color", c2: "colors.Color") -> bool:
    return c1.lhex == c2.lhex


def equal_hsl(c1: "colors.Color", c2: "colors.Color") -> bool:
    return colors.Color(c1, 1).csshsl == colors.Color(c2, 1).csshsl


def equal_hsla(c1: "colors.Color", c2: "colors.Color") -> bool:
    return c1.csshsl == c2.csshsl


def equal_hash(c1: "colors.Color", c2: "colors.Color") -> bool:
    return hash(c1) == hash(c2)


def get_cspace(color: Union[CTuple, "colors.Color"]):
    cspace = getattr(color, "cspace", "unknown")
    if isinstance(color, HSLTuple):
        cspace = "hsl"
    elif isinstance(color, RGBTuple):
        cspace = "rgb"
    elif isinstance(color, HSVTuple):
        cspace = "hsv"
    elif isinstance(color, HWBTuple):
        cspace = "hwb"
    elif isinstance(color, YUVTuple):
        cspace = "yuv"
    elif isinstance(color, XYZTuple):
        cspace = "xyz"
    elif isinstance(color, LabTuple):
        cspace = "lab"
    elif isinstance(color, LChTuple):
        cspace = "lch"
    elif isinstance(color, CMYKTuple):
        cspace = "cmyk"

    return cspace


def normalize_1base(c: CTuple) -> CTuple:
    cls = c.__class__
    return cls(*tuple(min(max(v, 0), 1) for v in c))  # type: ignore


def normalize_huebase(c: CTuple) -> CTuple:
    cls = c.__class__
    return cls(1 if c[0] == 1 else c[0] % 1, *tuple(min(max(v, 0), 1) for v in c[1:3]))  # type: ignore


def normalize_lab(c: CTuple) -> LabTuple:
    return LabTuple(
        min(max(c[0], 0), 400), *tuple(min(max(v, -160), 160) for v in c[1:3])
    )


def normalize_lch(c: CTuple) -> LChTuple:
    return LChTuple(
        min(max(c[0], 0), 400), min(abs(c[1]), 230), 1 if c[2] == 1 else c[2] % 1
    )


def normalize(color: CTuple, cspace: str = None) -> CTuple:
    cspace = cspace if cspace else get_cspace(color)

    if cspace in ["rgb", "yuv", "xyz", "cmyk"]:
        return normalize_1base(color)
    elif cspace == "lab":
        return normalize_lab(color)
    elif cspace == "lch":
        return normalize_lch(color)
    elif hueprop(cspace) == 0:
        return normalize_huebase(color)

    return color


def _apply_f(t: CTuple, f: float, op: Callable[[float, float], float]) -> CTuple:
    cls = t.__class__
    return cls(*tuple(op(v, f) for v in t))  # type: ignore


def _apply(color1: "colors.Color", color2: Union["colors.Color", CTuple, float], op: Callable[[float, float], float], cspace: str = None) -> "colors.Color":
    if cspace is None:
        if isinstance(color2, tuple):
            cspace = get_cspace(color2)
        else:
            cspace = color1.cspace

    ctup1 = getattr(color1, cspace)

    if isinstance(color2, (int, float)):
        return colors.Color(normalize(_apply_f(ctup1, color2, op)))

    ctup2 = getattr(color2, cspace) if isinstance(color2, colors.Color) else color2
    cls = ctup1.__class__

    ctup = normalize(cls(*tuple(op(p1, p2) for p1, p2 in zip(ctup1, ctup2))))
    return colors.Color(ctup)


def add(color1: "colors.Color", color2: Union["colors.Color", CTuple, float], cspace: str = None) -> "colors.Color":
    return _apply(color1, color2, operator.add, cspace)


def sub(color1: "colors.Color", color2: Union["colors.Color", CTuple, float], cspace: str = None) -> "colors.Color":
    return _apply(color1, color2, operator.sub, cspace)


def mul(color1: "colors.Color", color2: Union["colors.Color", CTuple, float], cspace: str = None) -> "colors.Color":
    return _apply(color1, color2, operator.mul, cspace)


def div(color1: "colors.Color", color2: Union["colors.Color", CTuple, float], cspace: str = None) -> "colors.Color":
    return _apply(color1, color2, operator.truediv, cspace)


def mix_linear(v1: float, v2: float, ratio: float = .5, gamma: float = 1):
    if gamma == 1:
        return (1 - ratio) * v1 + ratio * v2
    return ((1 - ratio) * v1 ** gamma + ratio * v2 ** gamma) ** (1 / gamma)


def mix_hue(v1: float, v2: float, ratio: float = .5, gamma: float = 1):
    if abs(v1 - v2) > .5:
        if v1 < v2:
            v1 += 1
        else:
            v2 += 1

    return mix_linear(v1, v2, ratio, gamma) % 1


def average(
    colorlist: List["colors.Color"], weights: List[float] = [], cspace: str = None
) -> "colors.Color":
    if len(colorlist) < 1:
        return colors.Color()
    elif len(colorlist) == 1:
        return colorlist[0]

    if cspace is None or cspace not in COLORSPACES:
        cspace = colorlist[0].cspace

    lenw, lenc = len(weights), len(colorlist)

    if lenw < lenc:
        weights = weights + [1 for _ in range(lenc - lenw)]
    elif lenw > lenc:
        weights = weights[:lenc]

    sw = sum(weights)
    weights = [w / sw for w in weights]

    cls = getattr(colorlist[0], cspace).__class__
    retp = cls()
    propnum = len(retp)
    alpha = 0

    for c, w in zip(colorlist, weights):
        cprop = getattr(c, cspace)
        retp = tuple(retp[i] + cprop[i] * w for i in range(propnum))
        alpha += c.alpha * w

    return colors.Color(cls(*retp), alpha=alpha, cspace=cspace)



import click
import repacolors
import repacolors.palette
import repacolors.schemes
import sys
import os
import subprocess  # nosec


def from_stdin():
    colordef = []
    # read from stdin
    for line in sys.stdin:
        colordef.append(line.strip())

    return colordef


def pick_external(picker="xcolor"):
    proc = subprocess.Popen(picker, stdout=subprocess.PIPE)  # nosec
    res = proc.communicate()[0].strip().decode()
    return repacolors.Color(res)


@click.group(invoke_without_command=False)
@click.pass_context
def color(ctx):
    pass


@color.command()
@click.argument("name", nargs=-1)
@click.option("--format", default="hexdisplay", help="Color format to show.")
def palette(name, format):
    """Get colors of given palette"""
    if len(name) == 0:
        names = repacolors.palette.PALETTES.keys()
        click.echo("List of available palette names:")
        click.echo(", ".join(names))
    else:
        colors = []
        for n in name:
            try:
                colors += repacolors.palette.get_palette(n)
            except KeyError:
                click.echo(f"Palette '{n}' not found.", err=True)

        for cdef in colors:
            c = repacolors.Color(cdef)
            c.print(format)


@color.command()
@click.argument("colordef", nargs=-1)
@click.option("--format", default="display", help="Color format to show.")
def display(colordef, format):
    """Display information about the provided colors."""
    if len(colordef) == 0:
        colordef = from_stdin()

    for cdef in colordef:
        c = repacolors.Color(cdef)
        c.print(format)


@color.command()
@click.argument("colordef", nargs=-1)
@click.option("--contrast", default=4.5, help="Required contrast")
@click.option("--format", default="display", help="Color format to show.")
@click.option("-v", "--verbose", "verbose", default=False, is_flag=True, help="Print verbose output.")
def adjust_contrast(colordef, contrast, format, verbose):
    """Adjust the colors to match required contrast ratio."""
    # check stdin for more colors
    if len(colordef) < 2 and not sys.stdin.isatty():
        stdincolors = from_stdin()
        colordef = (list(colordef) + list(stdincolors))[:2]

    if len(colordef) == 0:
        click.echo("At least one color definition required!")
        sys.exit(1)

    c1 = repacolors.Color(colordef[0])
    # still not enough colors, fall back to black/white
    if len(colordef) < 2:
        # using #757575 as middle point
        c2 = repacolors.Color("#fff") if c1.luminance < .178 else repacolors.Color("#000")
    else:
        c2 = repacolors.Color(colordef[1])

    adjc1, adjc2 = c1.adjust_contrast(c2, contrast)

    if not verbose or not sys.stdout.isatty():
        adjc1.print(format)
        adjc2.print(format)

    else:
        if c1 == adjc1 and c2 == adjc2:
            print(f"Contrast was already OK! ({c1.contrast_ratio(c2)} > {contrast})")
            print(f"{adjc2.termfg}{adjc1.termbg}  {adjc1.lhex}  ")
            print(f"{adjc1.termfg}{adjc2.termbg}  {adjc2.lhex}  ")
        else:
            print(f"Colors adjusted. ({c1.contrast_ratio(c2):.4f} => {adjc1.contrast_ratio(adjc2):.4f})")
            print(f"{c2.termfg}{c1.termbg}  {c1.lhex}  {c2.termreset} => {adjc2.termfg}{adjc1.termbg}  {adjc1.lhex}  ")
            print(f"{c1.termfg}{c2.termbg}  {c2.lhex}  {c1.termreset} => {adjc1.termfg}{adjc2.termbg}  {adjc2.lhex}  ")

        print(c1.termreset)
        adjc1.print(format)
        adjc2.print(format)


@color.command()
@click.argument("name", nargs=-1)
def colorwheel(name):
    """Display colorwheel defined by `name` or created by the colors provided via stdin."""
    if len(name) == 0:
        colors = from_stdin()
        cscale = repacolors.ColorScale(colors)
        cwheel = repacolors.ColorWheel(cscale)
    else:
        n = name[0].lower()
        if n in ["rgb", "hsl"]:
            cwheel = repacolors.schemes.RGB
        elif n in ["lab", "lch"]:
            cwheel = repacolors.schemes.LAB
        else:
            cwheel = repacolors.schemes.RYB

    cwheel.print()


@color.command()
@click.option("--format", default="display", help="Color format to show.")
@click.option("-n", "--number", "number", default=1, help="Number of colors to pick.")
@click.option("-c", "--copy", "copy", default=False, is_flag=True, help="Copy to clipboard (using `xsel`)")
def pick(format, number, copy):
    """Pick colors from your desktop."""
    # TODO copy to clipboard
    # TODO draw color wheel

    colorpicker = os.environ.get("COLORPICKER")
    colors = []

    # if color picker is not set, try to use integrated one
    if not colorpicker:
        try:
            import repacolors.picker
            colors = repacolors.picker.pick(number)
        except ImportError:
            pass

    while number > len(colors):
        colors.append(pick_external())

    for c in colors:
        c.print(format)

    if copy:
        try:
            proc = subprocess.Popen(["xsel", "-p"], stdin=subprocess.PIPE)  # nosec
            proc.stdin.write(c.lhex.encode())
            proc.stdin.close()
            proc.wait()
        except subprocess.SubProcessError:
            pass


@color.command()
@click.argument("colors", nargs=-1)
@click.option("--format", default="display", help="Scale format to show.")
@click.option("-s", "--steps", "steps", default=None, type=int, help="Number of steps.")
def scale(colors, format, steps):
    """Display color scale defined by the colors provided."""
    if len(colors) == 0:
        colors = from_stdin()
    cscale = repacolors.scale.ColorScale(colors)
    cscale.print(fmt=format, steps=steps)


if __name__ == "__main__":
    color()



from .colors import Color
from .types import RGBTuple
from .scale import ColorScale, project_domain, linear_ip_f
from typing import Union, List
import math


def _cubehelix(pos: float, start: float = 0, rotations: float = -1.5, lightness: float = None, hue: float = None):
    if lightness is None:
        lightness = pos

    if hue is None:
        hue = 1

    alpha = math.tau * ((start + .3333) + (rotations * pos))
    amp = (hue * lightness * (1 - lightness)) / 2
    cos_a = math.cos(alpha)
    sin_a = math.sin(alpha)
    r = lightness + amp * ((-0.14861 * cos_a) + (1.78277 * sin_a))
    g = lightness + amp * ((-0.29227 * cos_a) - (0.90649 * sin_a))
    b = lightness + amp * (1.97294 * cos_a)

    return Color(RGBTuple(r, g, b))


class CubeHelix(ColorScale):
    """CubeHelix color scale

    Dave Green's [cubehelix color scheme](http://www.mrao.cam.ac.uk/~dag/CUBEHELIX/)
    """

    def __init__(
        self,
        start: float = 0,
        rotations: float = -1.5,
        lightness: Union[float, List[float]] = [0, 1],
        hue: Union[float, List[float]] = [1, 1],
        gamma_correction: float = 1.0,
        domain: List[float] = None,
        gamma: float = 1.0,
        luminance_map: List[float] = [0, 100],
        name: str = "cubehelix"
    ):
        super().__init__(
            [Color("#000"), Color("#fff")],
            domain=domain,
            gamma=gamma,
            cspace="lab",
            gamma_correction=gamma_correction,
            interpolator="linear",
            luminance_map=luminance_map,
            name=name
        )

        self.start = start
        self.rotations = rotations

        self.hue = hue
        self.lightness = lightness

    @property
    def hue(self):
        return self._hue

    @hue.setter
    def hue(self, hue: Union[float, List[float]]):
        if isinstance(hue, (float, int)):
            hue = [hue, hue]
        self._hue = hue

    @property
    def lightness(self):
        return [lumin / 100 for lumin in self.luminance_map]

    @lightness.setter
    def lightness(self, lightness: Union[float, List[float]]):
        if isinstance(lightness, (float, int)):
            lightness = [lightness, lightness]
        self.luminance_map = [light * 100 for light in lightness]

    def _get_color_for_pos(self, pos: float) -> Color:
        projpos = project_domain(pos, self.domain)

        if self.gamma != 1.0:
            projpos = projpos ** self.gamma

        lightness = (linear_ip_f(self.luminance_map, projpos) / 100) ** self.gamma_correction
        hue = linear_ip_f(self.hue, projpos)

        return _cubehelix(
            projpos,
            start=self.start,
            rotations=self.rotations,
            lightness=lightness,
            hue=hue
        )



"""Named tuples for color spaces
"""

from typing import Tuple, NamedTuple, Optional

CTuple = Tuple[float, ...]


class HSLTuple(NamedTuple):
    hue: float = 0
    saturation: float = 0
    lightness: float = 0


class RGBTuple(NamedTuple):
    red: float = 0
    green: float = 0
    blue: float = 0


class HSVTuple(NamedTuple):
    hue: float = 0
    saturation: float = 0
    value: float = 0


class HWBTuple(NamedTuple):
    hue: float = 0
    whiteness: float = 0
    blackness: float = 0


class YUVTuple(NamedTuple):
    y: float = 0
    u: float = 0
    v: float = 0


class XYZTuple(NamedTuple):
    x: float = 0
    y: float = 0
    z: float = 0


class LabTuple(NamedTuple):
    l: float = 0
    a: float = 0
    b: float = 0


class LChTuple(NamedTuple):
    l: float = 0
    c: float = 0
    h: float = 0


class CMYKTuple(NamedTuple):
    c: float = 0
    m: float = 0
    y: float = 0
    k: float = 0


COLORSPACES = {
    "rgb": RGBTuple,
    "hsl": HSLTuple,
    "hsv": HSVTuple,
    "hwb": HWBTuple,
    "lab": LabTuple,
    "lch": LChTuple,
    "xyz": XYZTuple,
    "yuv": YUVTuple,
    "cmyk": CMYKTuple,
}


def hueprop(cspace: str) -> Optional[int]:
    if 'h' in cspace:
        return cspace.index('h')
    return None



"""
https://en.wikipedia.org/wiki/Blend_modes
"""
from . import colors
from .types import RGBTuple


def _normal(fgv: float, bgv: float) -> float:
    return fgv


def _multiply(fgv: float, bgv: float) -> float:
    return fgv * bgv


def _screen(fgv: float, bgv: float) -> float:
    return 1 - (1 - fgv) * (1 - bgv)


def _overlay(fgv: float, bgv: float) -> float:
    return 2 * fgv * bgv if bgv < 0.5 else 1 - 2 * (1 - fgv) * (1 - bgv)


def _hardlight(fgv: float, bgv: float) -> float:
    return _overlay(bgv, fgv)


def _softlight(fgv: float, bgv: float) -> float:
    if fgv <= .5:
        return bgv - (1 - 2 * fgv) * bgv * (1 - bgv)

    if bgv <= .25:
        return bgv + (2 * fgv - 1) * (((16 * bgv - 12) * bgv + 3) * bgv)

    return bgv + (2 * fgv - 1) * (bgv ** 2 - bgv)


BLEND_MODES = {
    "normal": _normal,
    "multiply": _multiply,
    "screen": _screen,
    "overlay": _overlay,
    "hardlight": _hardlight,
    "hard-light": _hardlight,
    "softlight": _softlight,
    "soft-light": _softlight,
    # ...
}


def blend(fg: "colors.Color", bg: "colors.Color", mode: str = "normal", gamma: float = None) -> "colors.Color":
    if mode not in BLEND_MODES:
        mode = "normal"

    if mode == "normal":
        if fg.alpha == 1:
            return fg

        if fg.alpha == 0:
            return bg

    if gamma is None:
        if mode == "normal":
            gamma = 2.2
        else:
            gamma = 1.0

    blendfn = BLEND_MODES[mode]

    blended = RGBTuple(
        *tuple(
            (fg.alpha * blendfn(fgv, bgv) ** gamma + (1 - fg.alpha) * bgv ** gamma) ** (1 / gamma)
            for fgv, bgv in zip(fg.rgb, bg.rgb)
        )
    )

    return colors.Color(blended, bg.alpha)



from contextlib import contextmanager, ContextDecorator
import math
import xcffib
import xcffib.render
from xcffib.xproto import WindowClass, CW, EventMask, InputFocus, Time, GrabMode, Atom, ImageFormat
from time import sleep
from .convert import rgb2hsl, hsl2rgb
from .colors import Color

MOUSE_BUTTON_LEFT = 1
MOUSE_BUTTON_RIGHT = 3


def x8to16(i):
    return 0xffff * (i & 0xff) // 0xff


def xcolor(color):
    return (x8to16(i) for i in color + (0xffff,))


def border_color(rgb):
    hsl = rgb2hsl(tuple(x / 255 for x in rgb))
    if hsl.lightness < .5:
        # dark => lighten
        color = tuple(int(255 * x) for x in hsl2rgb((hsl.hue, hsl.saturation, hsl.lightness + .25)))
    else:
        # light => darken
        color = tuple(int(255 * x) for x in hsl2rgb((hsl.hue, hsl.saturation, hsl.lightness - .25)))

    return color


@contextmanager
def create_font_cursor(conn, which):
    """create mouse cursor

    Params:
    - conn: xcb connection
    - which: cursor index

    Yields:
    - the cursor object (id)
    """

    black, white = (0, 0, 0), (65535, 65535, 65535)
    font, cursor = None, None
    fname = "cursor"

    try:
        font = conn.generate_id()
        cursor = conn.generate_id()
        conn.core.OpenFont(font, len(fname), fname)
        conn.core.CreateGlyphCursor(cursor, font, font, which, which + 1, *black, *white)

        yield cursor
    finally:
        if cursor:
            conn.core.FreeCursor(cursor)
        if font:
            conn.core.CloseFont(font)


@contextmanager
def pick_coord(conn, cursor=None):
    """Changes the cursor and grabs the pointer

    Params:
    - conn: xcb connection
    - cursor: cursor object
    """
    setup = conn.get_setup()
    try:
        conn.core.GrabPointer(
            0, setup.roots[0].root,
            EventMask.PointerMotion | EventMask.ButtonPress,
            GrabMode.Async, GrabMode.Async,
            Atom._None,
            cursor or Atom._None,
            Atom._None
        )
        conn.flush()
        yield
    finally:
        conn.core.UngrabPointer(Atom._None)


def events(conn, sleep_time=0.001):
    """Return a generator that yields the incoming X events"""
    while True:
        event = conn.poll_for_event()
        if event:
            yield event
        else:
            sleep(sleep_time)


def get_pointer_position(conn, event=None):
    """Get position of cursor pointer from the event,
    falls back to `query_pointer`.

    Params:
    - conn: xcb connection
    - event: xcb event

    Returns:
    - (tuple of int): (x, y) coordinates
    """
    try:
        return event.root_x, event.root_y
    except AttributeError:
        setup = conn.get_setup()
        ppos = conn.core.QueryPointer(setup.roots[0].root).reply()
        return ppos.root_x, ppos.root_y


# TODO get 3x3 or 5x5 pixels to easier picking
def get_pixel(conn, x, y):
    """Get the pixel color at the given position

    Params:
    - conn: xcb connection
    - x, y: coordinates

    Returns:
    - (tuple of int): (r, g, b) color tuple
    """

    setup = conn.get_setup()
    reply = conn.core.GetImage(ImageFormat.ZPixmap, setup.roots[0].root, x, y, 1, 1, 0xffffffff).reply()
    # format: BGRX
    image_data = reply.data.buf()

    return (image_data[2], image_data[1], image_data[0])


class ColorPreview:
    """Small window to disp a picked color

    Params:
    - margin: distance to the pointer
    - size: size of the preview
    - border: border size
    """

    def __init__(self, conn, margin, size, border, colors=None, length=1):
        self._colors = colors if colors is not None else [(0, 0, 0)] * length
        self._length = len(self._colors)
        self._connection = conn
        self._setup = conn.get_setup()
        self._render = conn(xcffib.render.key)
        self._root = self._setup.roots[0]
        self.margin = margin
        self.size = size
        self.border = border
        self._distance_x = margin + int(size / 2) + int(size * self._length)
        self._distance_y = margin + int(size / 2)
        self._window = None
        self._pid = None
        self._format = None
        self._picked = 0
        q = self._connection.core.GetGeometry(self._root.root)
        res = q.reply()
        self.screen_width, self.screen_height = res.width, res.height

    def _find_format(self):
        cookie = self._render.QueryPictFormats()
        reply = cookie.reply()
        for depth in reply.screens[0].depths:
            if depth.depth == self._root.root_depth:
                for vis in depth.visuals:
                    if vis.visual == self._root.root_visual:
                        return vis.format

    def map(self):
        """Creates and maps the window & gc

        Returns:
        - xcb window
        """
        if not self._window:
            self._window = self._connection.generate_id()
            self._pid = self._connection.generate_id()
            self._format = self._find_format()
            self._connection.core.CreateWindow(
                self._root.root_depth, self._window, self._root.root,
                0, 0, self.size * len(self._colors), self.size, 0,
                WindowClass.InputOutput,
                self._root.root_visual,
                CW.BackPixel | CW.OverrideRedirect | CW.EventMask,
                [self._root.white_pixel, True, EventMask.ButtonPress | EventMask.Exposure | EventMask.PointerMotion])
        self._render.CreatePicture(self._pid, self._window, self._format, 0, [])
        self._connection.core.MapWindow(self._window)
        self._connection.flush()
        return self._window

    def unmap(self):
        """Unmaps the window"""
        if self._window:
            self._connection.core.UnmapWindow(self._window)
            self._connection.core.DestroyWindow(self._window)
            self._window = None
            self._connection.flush()

    def move(self, x, y):
        """Moves the windowdow based on the given (mouse) position
        and tries to keep it visible.

        Param:
        - x, y: cursor position
        """
        x_center = (x - self.screen_width / 2)
        y_center = -(y - self.screen_height / 2)
        rad = math.atan2(-y_center, -x_center)
        offset_y = int(self._distance_y * -math.sin(rad))
        offset_x = int(self._distance_x * math.cos(rad))

        self._connection.core.ConfigureWindow(self._window, xcffib.xproto.ConfigWindow.X | xcffib.xproto.ConfigWindow.Y, [x - int(self.size / 2) + offset_x, y - int(self.size / 2) + offset_y])

    def _draw_rectangle(self, x, y, width, height, color=None):
        """Draws a rectangle with the current foreground color on the window.

        Params:
        - x/y: position
        - width/height: size
        """
        if color is None:
            color = xcolor(self._colors[0])
        if isinstance(color, tuple):
            color = xcolor(color)
        self._render.FillRectangles(
            xcffib.render.PictOp.Src,
            self._pid, color, 1,
            [xcffib.xproto.RECTANGLE.synthetic(x, y, width, height)]
        )

    def _draw_border(self, rgb=(255, 255, 255), i=0):
        """Drows border around the window.

        Params:
        - r/g/b: red/green/blue
        """
        color = border_color(rgb)

        for rect in (
                (self.size * i, 0, self.border, self.size),
                (self.size * i, 0, self.size, self.border),
                (self.size * (i+1) - self.border, 0, self.size, self.size),
                (self.size * i, self.size - self.border, self.size, self.size)):
            self._draw_rectangle(*rect, xcolor(color))

    def draw(self, rgb=None):
        """Redraws the window.

        Param:
        - rgb (tuple of int): color
        """
        rgb = rgb or self._colors[0]
        self._colors[0] = rgb
        for i, color in enumerate(self._colors):
            self._draw_rectangle(self.size * i, 0, self.size, self.size, color)
            self._draw_border(color, i)
        self._connection.flush()

    def pick(self, rgb):
        self._picked += 1

        if self._picked >= self._length:
            return True

        self._colors = [rgb] + self._colors[:-1]


class WindowMapper(ContextDecorator):
    """Maps all the windows, ensures unmapping"""

    def __init__(self, window):
        super().__init__()
        self._mapped = False
        self._window = window

    def __enter__(self):
        self._window.map()
        self._mapped = True
        return self

    def __exit__(self, *args):
        self._window.unmap()


def pick(length=1):
    conn = xcffib.connect()
    wnd = ColorPreview(conn, 30, 20, 1, length=length)

    with WindowMapper(wnd) as mapper,\
            create_font_cursor(conn, 34) as cursor,\
            pick_coord(conn, cursor):
        for event in events(conn):
            x, y = get_pointer_position(conn, event)
            rgb = get_pixel(conn, x, y)
            if isinstance(event, (xcffib.xproto.ExposeEvent, xcffib.xproto.MotionNotifyEvent)):
                wnd.draw(rgb)
                wnd.move(x, y)
            elif isinstance(event, xcffib.xproto.ButtonPressEvent):
                if event.detail == MOUSE_BUTTON_RIGHT:
                    # quit
                    break
                elif event.detail == MOUSE_BUTTON_LEFT:
                    # pick color
                    should_quit = wnd.pick(rgb)
                    wnd.draw(rgb)
                    if should_quit:
                        break

    return [Color(rgb) for rgb in reversed(wnd._colors)]



import json
import os
import sys
import math
import re
from itertools import zip_longest
from typing import Dict, Any, Optional, Callable, Iterator, List, Union, Tuple
from . import convert
from . import terminal
from . import distance
from . import blend
from . import ops
from .types import *

DIR = os.path.dirname(os.path.realpath(__file__))
CSSJSON = os.path.join(DIR, "css-color-names.json")

with open(CSSJSON, "r") as f:
    CSSNAME2HEX = json.load(f)

HEX2CSSNAME = dict((hx, name) for name, hx in CSSNAME2HEX.items())


def name2hex(name: str) -> str:
    return CSSNAME2HEX.get(name, None)


def hex2name(hx: str) -> str:
    return HEX2CSSNAME.get(hx, None)


# TODO
COLORCACHE: Dict[str, Dict[str, Any]] = {}


# TODO
def get_color(hx, cspace="rgb", container=COLORCACHE):
    if hx in container:
        color = container[hx]
    else:
        color = {"hex": hx, "rgb": convert.hex2rgb(hx), "name": hex2name(hx)}
        container[hx] = color

    if cspace not in color and hasattr(convert, f"rgb2{cspace}"):
        color[cspace] = getattr(convert, f"rgb2{cspace}")(RGBTuple(*color["rgb"]))

    return color


# TODO
def closest(col: CTuple, n: int = 3, cspace: str = "rgb"):
    named_colors = HEX2CSSNAME.keys()
    chx = convert.rgb2hex(RGBTuple(*col), True)
    if cspace != "rgb":
        col = getattr(convert, f"rgb2{cspace}")(*col)
    closests = []

    for nch in named_colors:
        if chx == nch:
            continue
        nc = get_color(nch, cspace)
        distance = distance.distance(col, nc[cspace])
        closests.append({"color": nc, "distance": distance})
        # not the most efficient way, but...
        closests.sort(key=lambda c: c["distance"])
        closests = closests[:n]

    return closests


class ColorProperty:
    def __init__(self, name, mainprop):
        self.name = name
        self.mainprop = mainprop

    def __get__(self, obj: "Color", objtype: type):
        prop = getattr(obj, self.mainprop)
        return getattr(prop, self.name)

    def __set__(self, obj: "Color", val: Any):
        if not obj._initialized:
            prop = getattr(obj, self.mainprop)
            proplist = list(prop)
            proptype = prop.__class__
            idx = prop._fields.index(self.name)
            proplist[idx] = val
            setattr(obj, self.mainprop, proptype(*proplist))
        else:
            raise TypeError("Should not modify an existing 'Color' instance")


class ColorSpaceProperty:
    def __init__(self, name: str = "lab"):
        self.name = name
        self.privname = "_" + self.name

    def __get__(self, obj: "Color", objtype: type):
        privattr = getattr(obj, self.privname, None)

        # already cached
        if privattr is not None:
            return privattr

        # not cached yet, check for convert function
        convertfn = getattr(convert, f"rgb2{self.name}", None)
        if convertfn:
            privattr = convertfn(obj._rgb)
            setattr(obj, self.privname, privattr)
            return privattr

        # cannot convert
        raise TypeError(f"Cannot convert 'Color' to {self.name} format")

    def __set__(self, obj: "Color", val: Any):
        if not obj._initialized:
            # check for convert function
            convertfn = getattr(convert, f"{self.name}2rgb", None)
            if convertfn:
                rgb = convertfn(val)
                obj.rgb = rgb
                setattr(obj, self.privname, val)
            else:
                # no conversion function found
                raise TypeError(f"Cannot convert {self.name} format to 'Color'")
        else:
            raise TypeError("Should not modify an existing 'Color' instance")


class Color(terminal.TerminalColor):
    """Color object

    Can be initialized with many kind of color definitions, and output different
    formats as well.
    """

    DISPLAY_HEIGHT = 12
    DISPLAY_WIDTH = 12
    DISPLAY_BORDER = 2

    red = ColorProperty("red", "rgb")
    green = ColorProperty("green", "rgb")
    blue = ColorProperty("blue", "rgb")
    r = ColorProperty("red", "rgb256")
    g = ColorProperty("green", "rgb256")
    b = ColorProperty("blue", "rgb256")
    hue = ColorProperty("hue", "hsl")
    saturation = ColorProperty("saturation", "hsl")
    lightness = ColorProperty("lightness", "hsl")
    whiteness = ColorProperty("whiteness", "hwb")
    blackness = ColorProperty("blackness", "hwb")
    value = ColorProperty("value", "hsv")
    cie_l = ColorProperty("l", "lab")
    cie_a = ColorProperty("a", "lab")
    cie_b = ColorProperty("b", "lab")
    cie_c = ColorProperty("c", "lch")
    cie_h = ColorProperty("h", "lch")
    cie_x = ColorProperty("x", "xyz")
    cie_y = ColorProperty("y", "xyz")
    cie_z = ColorProperty("z", "xyz")
    y = ColorProperty("y", "yuv")
    u = ColorProperty("u", "yuv")
    v = ColorProperty("v", "yuv")

    hex = ColorSpaceProperty("hex")
    hsv = ColorSpaceProperty("hsv")
    hwb = ColorSpaceProperty("hwb")
    ansi = ColorSpaceProperty("ansi")
    xyz = ColorSpaceProperty("xyz")
    lab = ColorSpaceProperty("lab")
    lch = ColorSpaceProperty("lch")
    yuv = ColorSpaceProperty("yuv")
    hsv = ColorSpaceProperty("hsv")
    yiq = ColorSpaceProperty("yiq")
    cmyk = ColorSpaceProperty("cmyk")

    def __init__(
        self,
        colordef: Any = None,
        alpha: float = None,
        equality: Optional[Callable[["Color", "Color"], bool]] = None,
        **kwargs,
    ):
        self._hsl = HSLTuple(0, 0, 0)
        self._rgb = RGBTuple(0, 0, 0)
        self._alpha = 1.0
        self.cspace = "hsl"

        self._initialized = False

        self._eq = equality if equality is not None else ops.equal_hex

        # from color
        if isinstance(colordef, Color):
            self._hsl = colordef._hsl
            self.rgb = colordef._rgb
            self.alpha = colordef.alpha
            self.cspace = colordef.cspace

        # from tuple
        elif isinstance(colordef, (tuple, list)):
            self._init_tuple(colordef)

        # from bytes
        elif isinstance(colordef, bytes):
            self._init_bytes(colordef)

        # from str
        elif isinstance(colordef, str):
            self._init_str(colordef)

        # override alpha
        if alpha is not None:
            self._alpha = alpha

        # additional parameters
        for k, v in kwargs.items():
            setattr(self, k, v)

        self._initialized = True

    def _init_tuple(self, colordef: Union[tuple, list]):
        for cspace, ctype in COLORSPACES.items():
            if isinstance(colordef, ctype):
                setattr(self, cspace, colordef)
                self.cspace = cspace
                break
        else:
            if len(colordef) >= 3:
                if abs(colordef[0]) > 1 or abs(colordef[1]) > 1 or abs(colordef[2]) > 1:
                    self.rgb256 = tuple(v % 256 for v in colordef[:3])
                else:
                    self.rgb = tuple(abs(v) for v in colordef[:3])
                self.cspace = "rgb"

                if len(colordef) >= 4:
                    self.alpha = colordef[3]

    def _init_bytes(self, colordef: bytes):
        if len(colordef) == 3:
            self.rgb256 = RGBTuple(*tuple(colordef))
            self.cspace = "rgb"
        elif len(colordef) == 4:
            self.rgb256 = RGBTuple(*tuple(colordef[:3]))
            self.alpha = list(colordef)[3] / 255
            self.cspace = "rgb"

    def _init_str(self, colordef: str):
        # from hex color
        if colordef.startswith("#"):
            self.hex = colordef
            self.cspace = "rgb"
            if len(colordef) == 5:
                self.alpha = int(colordef[4] * 2, 16) / 255
            elif len(colordef) == 9:
                self.alpha = int(colordef[7:9], 16) / 255
        elif (
            colordef.startswith("rgb")
            or colordef.startswith("hsl")
            or colordef.startswith("hwb")
            or colordef.startswith("lab")
        ):
            mode = colordef[:3]
            clr, self.alpha = self.parse_css_color_values(colordef, mode)
            setattr(self, mode, clr)
            self.cspace = mode
        elif colordef.startswith("gray"):
            begin = colordef.index("(")
            end = colordef.index(")")
            cnt = colordef[begin + 1 : end]

            cnt = re.sub(r"\s+", " ", cnt.replace("/", " ")).strip()
            values = [float(v.strip()) for v in cnt.split(" ")]

            self.lab = LabTuple(values[0], 0, 0)
            self.alpha = values[1] if len(values) > 1 else 1.0
        else:
            hx = name2hex(colordef)
            if hx:
                self.rgb = convert.hex2rgb(hx)
                self.cspace = "rgb"
                self._name = colordef
            else:
                self.rgb = Color._colorize(colordef)
                self.cspace = "rgb"

    @staticmethod
    def from_bytes(cbytes: bytes, byteorder: str = "RGB") -> "Color":
        red = 0
        green = 0
        blue = 0
        alpha = 1.0
        for i, component in enumerate(byteorder.upper()):
            if component == "R":
                red = cbytes[i]
            elif component == "G":
                green = cbytes[i]
            elif component == "B":
                blue = cbytes[i]
            elif component == "A":
                alpha = cbytes[i] / 255
        return Color(bytes([red, green, blue]), alpha=alpha)

    @staticmethod
    def parse_css_color_values(
        csscolordef: str, mode: str = "rgb"
    ) -> Tuple[CTuple, float]:
        """Parse CSS color definition

        https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
        """
        begin = csscolordef.index("(")
        end = csscolordef.index(")")
        cnt = csscolordef[begin + 1 : end]
        alpha = 1.0

        if "," in cnt:
            cnt = re.sub(r"\s+", " ", cnt).strip()
            values = [v.strip() for v in cnt.split(",")]
        else:
            cnt = cnt.replace("/", " ").strip()
            cnt = re.sub(r"\s+", " ", cnt)
            values = [v.strip() for v in cnt.split(" ")]

        if len(values) > 3:
            alpha = Color.parse_css_color_value(values[3], "alpha")

        cls = COLORSPACES.get(mode.lower(), RGBTuple)
        color = cls(*tuple(Color.parse_css_color_value(v, mode) for v in values[:3]))

        return color, alpha

    @staticmethod
    def parse_css_color_value(csscolorvalue: str, mode: str = "rgb") -> float:
        if csscolorvalue[-1] == "%":
            if mode.lower() == "lab":
                return float(csscolorvalue[:-1])
            return float(csscolorvalue[:-1]) / 100
        elif mode.lower() == "alpha":
            return float(csscolorvalue)
        elif mode.lower() in ["hsl", "hwb"]:
            # hue only
            if csscolorvalue[-4:] == "turn":
                return float(csscolorvalue[:-4])
            elif csscolorvalue[-3:] == "rad":
                return float(csscolorvalue[:-3]) / (2 * math.pi)
            elif csscolorvalue[-3:] == "deg":
                csscolorvalue = csscolorvalue[:-3]

            # degree
            return float(csscolorvalue) / 360
        elif mode.lower() == "lab":
            return float(csscolorvalue)

        # rgb
        return float(csscolorvalue) / 255

    @classmethod
    def _colorize(cls, obj: Any) -> RGBTuple:
        if isinstance(obj, RGBTuple):
            return obj

        if isinstance(obj, cls):
            return obj.rgb

        for cspace, ctype in COLORSPACES.items():
            if isinstance(obj, ctype):
                converter = getattr(convert, f"{cspace}2rgb", None)
                if converter:
                    return converter(obj)

        if isinstance(obj, tuple):
            if (
                isinstance(obj[0], (int, float))
                and isinstance(obj[1], (int, float))
                and isinstance(obj[2], (int, float))
            ):
                if abs(obj[0]) > 1 or abs(obj[1]) > 1 or abs(obj[2]) > 1:
                    return RGBTuple(*tuple((abs(o) % 256) / 255 for o in obj))
                else:
                    return RGBTuple(*tuple(abs(o) for o in obj))

        # use the hash
        try:
            hsh = hash(obj)
        except TypeError:
            hsh = hash(str(obj))

        r = hsh % 1e3
        g = int(hsh / 1e3) % 1e3
        b = int(hsh / 1e6) % 1e3

        return RGBTuple(r / 999, g / 999, b / 999)

    @classmethod
    def colorize(cls, obj: Any) -> "Color":
        if isinstance(obj, Color):
            return Color(obj)

        return Color(Color._colorize(obj))

    def set(self, **kwargs):
        return Color(self, **kwargs)

    def lighten(self, amount=10):
        return self.set(cie_l=self.cie_l + amount)

    def darken(self, amount=10):
        return self.set(cie_l=self.cie_l - amount)

    def saturate(self, amount=10):
        return self.set(cie_c=self.cie_c + amount)

    def desaturate(self, amount=10):
        return self.set(cie_c=max(0, self.cie_c - amount))

    def rotate(self, amount=0.1):
        return self.set(cie_h=self.cie_h + amount)

    def gray(self):
        return Color(LabTuple(self.cie_l, 0, 0))

    def mix(self, color: "Color", ratio: float = 0.5, cspace: str = None, gamma: float = None) -> "Color":
        if cspace is None or cspace not in COLORSPACES:
            cspace = self.cspace

        if gamma is None:
            gamma = 1.0

        ratio = min(abs(ratio), 1)
        prop1, prop2 = getattr(self, cspace), getattr(color, cspace)

        huep = hueprop(cspace)
        newprop = prop1.__class__(
            *tuple(
                ops.mix_linear(v1, v2, ratio, gamma)
                if idx != huep
                else ops.mix_hue(v1, v2, ratio, gamma)
                for v1, v2, idx in zip(prop1, prop2, range(len(prop1)))
            )
        )
        alpha = self.alpha * (1 - ratio) + color.alpha * ratio
        return Color(newprop, alpha=alpha, cspace=self.cspace)

    def closest_named(self, num: int = 3) -> List["Color"]:
        # TODO return closest named colors
        return []

    def distance(self, other: "Color") -> float:
        return distance.distance_cie94(self.lab, other.lab)

    def complementary(self, cspace: "str" = None):
        if cspace is None:
            cspace = self.cspace

        return ops.sub(Color("#fff"), self, cspace)

    def _divide_wheel(self, n: int, cspace: "str" = None):
        if cspace is None:
            cspace = self.cspace

        if cspace in ["rgb", "hsl", "hsv", "hwb"]:
            return [self.set(hue=self.hue + v / n) for v in range(n)]

        if cspace in ["lab", "lch", "xyz"]:
            return [self.set(cie_h=self.cie_h + v / n) for v in range(n)]

    def triad(self, cspace: "str" = None):
        return self._divide_wheel(3, cspace)

    def square(self, cspace: "str" = None):
        return self._divide_wheel(4, cspace)

    def tetrad(self, cspace: "str" = None):
        colors = self._divide_wheel(6, cspace)
        return colors[0:2] + colors[3:5]

    def split_complementary(self, cspace: "str" = None):
        colors = self._divide_wheel(12, cspace)
        return colors[0:1] + colors[5:6] + colors[7:8]

    def analogous(self, cspace: "str" = None):
        colors = self._divide_wheel(12, cspace)
        return [colors[-1]] + colors[0:2]

    def similar(self, other: "Color") -> bool:
        return self.distance(other) < 2.3

    @property
    def name(self):
        if getattr(self, "_name", None) is None:
            self._name = hex2name(self.lhex) or self.hex
        return self._name

    @property
    def rgb(self):
        return ops.normalize_1base(self._rgb)

    @rgb.setter
    def rgb(self, rgb: CTuple):
        if not self._initialized:
            self._rgb = RGBTuple(*rgb)  # NOT normalized
            hue = self.hue
            self._hsl = ops.normalize_huebase(convert.rgb2hsl(ops.normalize_1base(self._rgb)))  # type: ignore
            if self._hsl.saturation == 0:
                self._hsl = HSLTuple(hue, self._hsl.saturation, self._hsl.lightness)
        else:
            raise TypeError("Should not modify an existing 'Color' instance")

    @property
    def clipped(self):
        return self.rgb != self._rgb

    @property
    def rgb256(self):
        if getattr(self, "_rgb256", None) is None:
            hx = convert.rgb2hex(self.rgb, True)[1:]
            self._rgb256 = RGBTuple(
                *tuple(int(hx[v * 2 : v * 2 + 2], 16) for v in range(3))
            )
        return self._rgb256

    @rgb256.setter
    def rgb256(self, rgb: CTuple):
        self.rgb = ops.normalize_1base(RGBTuple(*tuple(c / 255 for c in rgb)))
        self._rgb256 = RGBTuple(*rgb)

    @property
    def hsl(self):
        return self._hsl

    @hsl.setter
    def hsl(self, hsl: CTuple):
        if not self._initialized:
            self._hsl = ops.normalize_huebase(HSLTuple(*hsl))  # type: ignore
            self._rgb = convert.hsl2rgb(self._hsl)
        else:
            raise TypeError("Should not modify an existing 'Color' instance")

    @property
    def luminance(self):
        # should be the same as y in 'xyz'
        if getattr(self, "_luminance", None) is None:
            rgb = self.rgb
            rgb_lum = tuple(
                c / 12.92 if c <= 0.03928 else ((c + 0.055) / 1.055) ** 2.4 for c in rgb
            )
            self._luminance = (
                0.2126 * rgb_lum[0] + 0.7152 * rgb_lum[1] + 0.0722 * rgb_lum[2]
            )

        return self._luminance

    @property
    def lhexa(self):
        """Hexadecimal notation with alpha
        """
        ha = f"{int(self.alpha * 255):02x}"
        return self.lhex + ha

    @property
    def hexa(self):
        ha = f"{int(self.alpha * 255):02x}"
        if ha[0] == ha[1] and len(self.hex) == 4:
            return self.hex + ha[0]

        return self.lhex + ha

    @property
    def lhex(self):
        if getattr(self, "_lhex", None) is None:
            self._lhex = convert.rgb2hex(self.rgb, True)
        return self._lhex

    @property
    def cssrgb(self):
        if getattr(self, "_cssrgb", None) is None:
            rgb256 = self.rgb256
            if self.alpha == 1:
                self._cssrgb = f"rgb({rgb256.red}, {rgb256.green}, {rgb256.blue})"
            else:
                self._cssrgb = f"rgba({rgb256.red}, {rgb256.green}, {rgb256.blue}, {self.alpha:.5g})"

        return self._cssrgb

    @property
    def csshsl(self):
        if getattr(self, "_csshsl", None) is None:
            sat = int(self.saturation * 1000) / 10
            lig = int(self.lightness * 1000) / 10
            if self.alpha == 1:
                self._csshsl = f"hsl({int(360 * self.hue)}, {sat:.4g}%, {lig:.4g}%)"
            else:
                self._csshsl = f"hsla({int(360 * self.hue)}, {sat:.4g}%, {lig:.4g}%, {self.alpha:.5g})"
        return self._csshsl

    @property
    def csshwb(self):
        if getattr(self, "_csshwb", None) is None:
            white = int(self.whiteness * 1000) / 10
            black = int(self.blackness * 1000) / 10
            if self.alpha == 1:
                self._csshwb = f"hwb({int(360 * self.hue)}deg {white:.4g}% {black:.4g}%)"
            else:
                self._csshwb = f"hwb({int(360 * self.hue)}deg {white:.4g}% {black:.4g}% / {self.alpha:.5g})"

        return self._csshwb

    @property
    def csslab(self):
        if getattr(self, "_csslab", None) is None:
            ciel = int(self.cie_l * 100) / 100
            ciea = int(self.cie_a * 100) / 100
            cieb = int(self.cie_b * 100) / 100
            if self.alpha == 1:
                self._csslab = f"lab({ciel:.4g}% {ciea:.4g} {cieb:.4g})"
            else:
                self._csslab = f"lab({ciel:.4g}% {ciea:.4g} {cieb:.4g} / {self.alpha:.5g})"

        return self._csslab

    @property
    def csslch(self):
        if getattr(self, "_csslch", None) is None:
            ciel = int(self.cie_l * 100) / 100
            ciec = int(self.cie_c * 100) / 100
            if self.alpha == 1:
                self._csslch = f"lch({ciel:.4g}% {ciec:.4g} {int(360 * self.cie_h)}deg)"
            else:
                self._csslch = f"lch({ciel:.4g}% {ciec:.4g} {int(360 * self.cie_h)}deg / {self.alpha:.5g})"

        return self._csslch

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        if not self._initialized:
            alpha = abs(value)
            self._alpha = alpha if alpha <= 1 else (alpha % 256) / 255
        else:
            raise TypeError("Should not modify an existing 'Color' instance")

    @property
    def pltc(self):
        """matplotlib color tuple"""
        return self.rgb + (self.alpha,)

    def contrast_ratio(self, other: "Color", property: str = "cie_y") -> float:
        """WCAG relative contrast ratio

        property: cie_y or luminance should work

        contrast_ratio > 4.5:1 - AA
        contrast_ratio > 7:1 - AAA
        """
        l1, l2 = getattr(self, property), getattr(other, property)
        if l2 > l1:
            l1, l2 = l2, l1

        return (l1 + 0.05) / (l2 + 0.05)

    def adjust_contrast(self, other: "Color", limit: float = 4.5, epsilon: Optional[float] = None) -> Tuple["Color", "Color"]:
        # adjust contrast of the given colors to fulfill WCAG requirements
        # increases/decreases HSL lightness only
        # first tries to adjust the main color, if the other is too close then changes that too
        if self.contrast_ratio(other) >= limit:
            return (self, other)

        if epsilon is None:
            epsilon = limit / 1000

        lighter = self.cie_y > other.cie_y

        c1 = self
        c2 = other
        cr = c1.contrast_ratio(c2)
        if lighter:
            increment = epsilon
        else:
            increment = -epsilon

        # try to adjust c1
        while cr < limit and c1.lightness < .999 and c1.lightness > .001:
            c1 = c1.set(lightness = c1.lightness + increment)
            cr = c1.contrast_ratio(c2)

        if cr < limit:
            # c1 was not enough try to adjust c2 too
            while cr < limit and c2.lightness < .999 and c2.lightness > .001:
                c2 = c2.set(lightness = c2.lightness - increment)
                cr = c1.contrast_ratio(c2)

        return (c1, c2)

    def __str__(self):
        return self.hex

    def __repr__(self):
        return f"<Color {self.name} - {self.csshsl}>"

    def __eq__(self, other):
        if isinstance(other, Color):
            return self._eq(self, other)

        return NotImplemented

    def __add__(self, other):
        if isinstance(other, (Color, tuple, float, int)):
            return ops.add(self, other)

        raise TypeError(f"Cannot add '{type(other)}' to 'Color'")

    def __sub__(self, other):
        if isinstance(other, (Color, tuple, float, int)):
            return ops.sub(self, other)

        raise TypeError(f"Cannot substract '{type(other)}' from 'Color'")

    def __mul__(self, other):
        if isinstance(other, (Color, tuple, float, int)):
            return ops.mul(self, other)

        raise TypeError(f"Cannot multiply '{type(other)}' with 'Color'")

    def __truediv__(self, other):
        if isinstance(other, (Color, tuple, float, int)):
            return ops.div(self, other)

        raise TypeError(f"Cannot multiply '{type(other)}' with 'Color'")

    def __radd__(self, color):
        return self + color

    def __rmul__(self, color):
        return self * color

    def __hash__(self):
        hsl = self.hsl
        return int(
            int(hsl.hue * 3.6e3) * 1e9
            + int(hsl.saturation * 1e3) * 1e6
            + int(hsl.lightness * 1e3) * 1e3
            + int(self.alpha * 1e3)
        )

    @property
    def textcolor(self):
        if getattr(self, "_textcolor", None) is None:
            w = Color("#fff")
            b = Color("#000")
            ctrw = self.contrast_ratio(w)
            ctrb = self.contrast_ratio(b)
            if ctrw > ctrb:
                self._textcolor = w
            else:
                self._textcolor = b

        return self._textcolor

    @property
    def termbg(self):
        rgb = self.rgb256
        return f"\x1b[48;2;{rgb.red};{rgb.green};{rgb.blue}m"

    @property
    def termfg(self):
        rgb = self.rgb256
        return f"\x1b[38;2;{rgb.red};{rgb.green};{rgb.blue}m"

    def _displayimage(
        self,
        width: int = None,
        height: int = None,
        border: int = None,
        bgcolors: List["Color"] = None,
    ) -> List[List["Color"]]:
        if width is None:
            width = self.DISPLAY_WIDTH
        if height is None:
            height = self.DISPLAY_HEIGHT
        if border is None:
            border = self.DISPLAY_BORDER

        w = width + 2 * border
        h = height + 2 * border
        img = []

        if bgcolors is None:
            bgcolors = getattr(
                self, "bgcolors", [Color(LabTuple(95, 0, 0)), Color(LabTuple(65, 0, 0))]
            )

        bgl = len(bgcolors)

        for y in range(h):
            line = []
            for x in range(w):
                bgc = bgcolors[(y + x) % bgl]
                if x < border or y < border or x > w - border - 1 or y > h - border - 1:
                    line.append(bgc)
                elif y < w / 2:
                    line.append(blend.blend(self, bgc))
                elif x < w / 2:
                    line.append(blend.blend(self, bgc, gamma=1))
                else:
                    line.append(self.set(alpha=1))
            img.append(line)

        return img

    @property
    def displayimage(self) -> List[List["Color"]]:
        return self._displayimage()

    @property
    def termimage(self) -> str:
        return terminal.draw(self.displayimage)

    @property
    def info(self):
        info = [
            self.lhex if self.lhex == self.name else f"{self.name} - {self.lhex}",
            self.cssrgb,
            self.csshsl,
            self.csslab,
        ]

        return "\n".join(info) + "\n"

    @property
    def display(self):
        timg = self.termimage
        info = self.info
        output = ["\n"]
        for timgl, infol in zip_longest(timg.split("\n"), info.split("\n")):
            output.append("  ")
            output.append(timgl or "")
            output.append("  ")
            output.append(infol or "")
            output.append("  \n")

        return "".join(output)

    @property
    def hexdisplay(self):
        return f"{self.termbg}{self.textcolor.termfg} {self.lhex} {self.termreset}"

    def print(self, fmt: str = "display", force_ansi: bool = False, stream = sys.stdout):
        if not force_ansi and not sys.stdout.isatty() and fmt in ["display", "hexdisplay"]:
            fmt = "lhex"

        content = getattr(self, fmt, self.lhex)
        print(content, file=stream)



"""
`repacolors` provides classes, functions and a command line utility to convert
colors from one color space / format to other, create `mathplotlib` compatible
color scales and more.

.. include:: ./documentation.md
"""

from .colors import Color
from .scale import ColorScale
from .cubehelix import CubeHelix
from .schemes import ColorWheel

__version__ = "0.5.0"



"""Color distance functions
"""

import math
from .types import *

# weighting factors
Wgraphic = (1, 0.045, 0.015)
Wtextile = (2, 0.048, 0.014)


def distance(c1: CTuple, c2: CTuple) -> float:
    """Color distance - eucledian
    Can be used with RGB or CIE-Lab values (CIE76)
    https://en.wikipedia.org/wiki/Color_difference#CIELAB_%CE%94E*

    For Lab colors: diff ~ 2.3 - just noticeable difference
    """

    return ((c1[0] - c2[0]) ** 2 + (c1[1] - c2[1]) ** 2 + (c1[2] - c2[2]) ** 2) ** 0.5


def distance_cie94(lab1: LabTuple, lab2: LabTuple, w: Tuple[float, float, float] = Wgraphic) -> float:
    """Color distance - CIE94
    https://en.wikipedia.org/wiki/Color_difference#CIE94
    """

    dL = lab1.l - lab2.l
    c1 = (lab1.a ** 2 + lab1.b ** 2) ** .5
    c2 = (lab2.a ** 2 + lab2.b ** 2) ** .5
    dC = c1 - c2
    dH2 = (lab1.a - lab2.a) ** 2 + (lab1.b - lab2.b) ** 2 - dC ** 2
    dH = 0 if dH2 <= 0 else dH2 ** .5
    sC = 1 + w[1] * c1
    sH = 1 + w[2] * c1

    return ((dL / w[0]) ** 2 + (dC / sC) ** 2 + (dH / sH) ** 2) ** .5


def distance_hue(h1: float, h2: float) -> float:
    """HUE based distance
    distance > 0 -> "clockwise"
    distance < 0 -> "counter-clocwise"
    """
    dist = h2 - h1
    if abs(dist) > .5:
        dist = (-1 if dist > 0 else 1) * (1 - abs(dist))

    return dist



from collections import OrderedDict
from ..colors import Color

DEFAULT_NAMES = [
    "black", "red", "green", "brown", "blue", "magenta", "cyan", "lightgray",
    "darkgray", "lightred", "lightgreen", "yellow", "lightblue", "lightmagenta", "lightcyan", "white",
]

class Theme(OrderedDict):

    def __init__(self, values=None, names=None, aliases=None):
        self._aliases = aliases or {}

        if values is None:
            values = {}
        if isinstance(values, list):
            values = enumerate(values)

        # apply default aliases
        self.apply_ordered_aliases(DEFAULT_NAMES)

        if names is not None:
            self.apply_ordered_aliases(names)

        super().__init__(values)

        self._fgidx = 7
        self._bgidx = 0

    def apply_ordered_aliases(self, names):
        for i, namelist in enumerate(names):
            if not isinstance(namelist, list):
                namelist = [namelist]
            for name in namelist:
                self.alias(name, i)

    @property
    def fg(self):
        if "foreground" in self._aliases:
            return self.foreground
        return self[self._fgidx]

    @property
    def bg(self):
        if "background" in self._aliases:
            return self.background
        return self[self._bgidx]

    def color(self, attr):
        if isinstance(attr, int):
            return Color(self[attr])
        else:
            return Color(getattr(self, attr))

    def display(self):
        output = ["\n"]
        output.append(Color(self.bg).termbg + " " * 8)
        output.append(Color(self.fg).termbg + " " * 8)
        for j in range(2):
            output.append("\n")
            for i in range(8):
                output.append(Color(self[i + j * 8]).termbg + "  ")
        output.append("\n")

        return "".join(output)

    def print(self):
        print(self.display())

    def __getattr__(self, attr):
        attr = self._aliases.get(attr, attr)
        return self[attr]

    def __setattr__(self, attr, value):
        if attr[0] != "_":
            self[attr] = value
        else:
            super().__setattr__(attr, value)

    def alias(self, alias, key):
        self._aliases[alias] = key



from .palette import get_scale
from .scale import ColorScale
from .colors import Color
from .types import HSLTuple, LChTuple, LabTuple
from .blend import blend
from .distance import distance_hue
from . import terminal
from typing import Tuple, List
import math


class ColorWheel:
    """Color wheel
    Colors arranged in a circle, 0-12 range (like on a clock)

    Defaults to classic RYB color wheel
    """

    DISPLAY_BORDER = 1

    def __init__(self, scale: ColorScale = None, cspace: str = "rgb"):
        if scale is None:
            scale = get_scale("rybw3", cyclic=True)

        self.scale = scale
        self.cspace = cspace

        # make it cyclic
        if scale.colors[0] != scale.colors[-1]:
            scale.colors.append(scale.colors[0])

        # "around the clock"
        self.scale.domain = [0, 12]

    def __getitem__(self, pos: float) -> Color:
        if isinstance(pos, (float, int)):
            pos = pos % 12
            return self.scale[pos]

        return NotImplemented

    def _get_position(self, color: Color) -> float:
        mindist, minidx, closest = 100.0, 0, color
        colors = self.scale.colors[:-1]

        # search for most similar color (hue) in scale
        for i, col in enumerate(colors):
            dist = abs(distance_hue(color.hue, col.hue))
            if dist < mindist:
                mindist = dist
                minidx = i
                closest = col

        lscale = len(colors)
        pidx, nidx = (minidx - 1) % lscale, (minidx + 1) % lscale
        pcol, ncol = colors[pidx], colors[nidx]
        diff = distance_hue(closest.hue, color.hue)
        pdiff = distance_hue(pcol.hue, color.hue)
        ndiff = distance_hue(ncol.hue, color.hue)

        # different sign of diffs - in between somewhere
        if pdiff * diff < 0:
            idx = minidx - abs(diff / (pdiff - diff))
        elif ndiff * diff < 0:
            idx = minidx + abs(diff / (ndiff - diff))
        else:
            idx = minidx

        return 12 * idx / lscale

    def _adjust(self, color: Color, refcolor: Color) -> Color:
        if self.cspace in ["lab", "lch"]:
            return refcolor.set(cie_h=color.cie_h)

        return refcolor.set(hue=color.hue)

    def _complementary(self, pos: float) -> Tuple[Color, Color]:
        return self[pos], self[pos + 6]

    def _triad(self, pos: float) -> Tuple[Color, Color, Color]:
        return self[pos], self[pos + 4], self[pos + 8]

    def _square(self, pos: float) -> Tuple[Color, Color, Color, Color]:
        return self[pos], self[pos + 3], self[pos + 6], self[pos + 9]

    def _tetrad(self, pos: float) -> Tuple[Color, Color, Color, Color]:
        return self[pos], self[pos + 2], self[pos + 6], self[pos + 8]

    def _split_complementary(self, pos: float) -> Tuple[Color, Color, Color]:
        return self[pos], self[pos + 5], self[pos + 7]

    def _analogous(self, pos: float) -> Tuple[Color, Color, Color]:
        return self[pos], self[pos + 1], self[pos + 11]

    def complementary(self, color: Color, adjust: bool = True) -> Tuple[Color, ...]:
        """Complementary color
        """
        pos = self._get_position(color)
        colors = self._complementary(pos)
        if not adjust:
            return colors

        return tuple(self._adjust(c, color) for c in colors)

    def triad(self, color: Color, adjust: bool = True) -> Tuple[Color, ...]:
        """Triad color scheme
        """
        pos = self._get_position(color)
        colors = self._triad(pos)
        if not adjust:
            return colors

        return tuple(self._adjust(c, color) for c in colors)

    def square(self, color: Color, adjust: bool = True) -> Tuple[Color, ...]:
        """Square color scheme
        """
        pos = self._get_position(color)
        colors = self._square(pos)
        if not adjust:
            return colors

        return tuple(self._adjust(c, color) for c in colors)

    def tetrad(self, color: Color, adjust: bool = True) -> Tuple[Color, ...]:
        """Tetrad/rectangle color scheme
        """
        pos = self._get_position(color)
        colors = self._tetrad(pos)
        if not adjust:
            return colors

        return tuple(self._adjust(c, color) for c in colors)

    def split_complementary(self, color: Color, adjust: bool = True) -> Tuple[Color, ...]:
        """Split complementary color scheme
        """
        pos = self._get_position(color)
        colors = self._split_complementary(pos)
        if not adjust:
            return colors

        return tuple(self._adjust(c, color) for c in colors)

    def analogous(self, color: Color, adjust: bool = True) -> Tuple[Color, ...]:
        """Analogous color scheme
        """
        pos = self._get_position(color)
        colors = self._analogous(pos)
        if not adjust:
            return colors

        return tuple(self._adjust(c, color) for c in colors)

    def monochromatic(self, color: Color, n : int = 5) -> Tuple[Color, ...]:
        """Monochromatic color scheme
        """
        deltal = 1 / n
        diffl = color.lightness % deltal
        lightnesses = [diffl + deltal * i for i in range(n)]
        return tuple(color.set(lightness=lness) for lness in lightnesses)

    def _displayimage(self, width: int = None, border: int = None, bgcolors: List[Color] = None) -> List[List["Color"]]:
        if border is None:
            border = self.DISPLAY_BORDER
        if width is None:
            ts = terminal.termsize()
            width = min(ts[0], ts[1] * 2) - border * 2 - 6 # -6 for prompt

        w = width + 2 * border
        img = []

        if bgcolors is None:
            bgcolors = getattr(
                self, "bgcolors", [Color(LabTuple(35, 0, 0)), Color(LabTuple(5, 0, 0))]
            )

        bgl = len(bgcolors)

        for y in range(w):
            line = []
            for x in range(w):
                bgc = bgcolors[(y + x) % bgl]
                ro, ri = width / 2, width / 3
                rx, ry = w / 2 - x, w / 2 - y
                r = (rx ** 2 + ry ** 2) ** .5
                if r < ro:
                    alpha = math.atan2(ry, rx)
                    pos = 12 * alpha / math.tau - 3
                    alpha = 1
                    if r < ri:
                        alpha = 1 - (ri - r) / ri
                    line.append(blend(self[pos].set(alpha=alpha), bgc))
                else:
                    line.append(bgc)
            img.append(line)

        return img

    def print(
        self,
        width: int = None,
        border: int = None,
        bgcolors: List["Color"] = None,
    ):
        print(terminal.draw(self._displayimage(width, border, bgcolors)))


class HSLColorWheel(ColorWheel):
    """HSL/RGB color wheel
    """

    def __init__(self, lightness: float = .5, saturation: float = 1.0):
        self.lightness = lightness
        self.saturation = saturation

    def __getitem__(self, pos: float) -> Color:
        hue = pos / 12
        return Color(HSLTuple(hue, self.saturation, self.lightness))

    def _get_position(self, color: Color) -> float:
        self.saturation = color.saturation
        self.lightness = color.lightness
        return color.hue * 12


class LChColorWheel(ColorWheel):
    """LCh/Lab color wheel
    """

    def __init__(self, cie_l: float = 60, cie_c: float = 100):
        self.cie_l = cie_l
        self.cie_c = cie_c
        self.cspace = "lab"

    def __getitem__(self, pos: float) -> Color:
        cie_h = pos / 12
        return Color(LChTuple(self.cie_l, self.cie_c, cie_h))

    def _get_position(self, color: Color) -> float:
        self.cie_l = color.cie_l
        self.cie_c = color.cie_c
        return color.cie_h * 12


RYB = ColorWheel()
HSL = HSLColorWheel()
RGB = HSL
LCH = LChColorWheel()
LAB = LCH



from .colors import Color
from .types import LabTuple
from .blend import blend
from typing import List, Any, Tuple, Callable, Union
from . import terminal
from functools import wraps
import math
import sys


def _binomial(i: int, n: int) -> float:
    """Binomial coefficient
    """
    return math.factorial(n) / float(math.factorial(i) * math.factorial(n - i))


def _bernstein(t: float, i: int, n: int) -> float:
    """Bernstein polynom
    """
    return _binomial(i, n) * (t ** i) * ((1 - t) ** (n - i))


def _bezier(
    points: List[Tuple[float, ...]], t: float, gamma: float = 1.0
) -> Tuple[float, ...]:
    """Calculate coordinate of a point in the bezier curve
    """
    n = len(points) - 1
    val = [0.0 for _ in range(len(points[0]))]
    for i, pos in enumerate(points):
        bern = _bernstein(t, i, n)
        val = [v + (p * bern) ** gamma for v, p in zip(val, pos)]

    return tuple(val)


def bezier_ip(
    colors: List[Color], pos: float = 0.5, cspace: str = "lab", gamma: float = 1.0
) -> Color:
    ctup = _bezier([getattr(c, cspace) for c in colors], pos, gamma)
    alpha = _bezier([(c.alpha,) for c in colors], pos)[0]
    cargs = {cspace: ctup, "alpha": min(1.0, alpha)}

    return Color(**cargs)  # type: ignore


def linear_ip(
    colors: List[Color], pos: float = 0.5, cspace: str = "lab", gamma: float = 1.0
) -> Color:

    lenc = len(colors)
    idx = int(pos * (lenc - 1))
    cols = colors[idx:idx + 2]
    if len(cols) == 1:
        return cols[0]

    col1, col2 = cols
    ratio = (pos - idx / (lenc - 1)) * (lenc - 1)

    return col1.mix(col2, ratio=ratio, cspace=cspace, gamma=gamma)


def linear_ip_f(lst: List[float], pos: float = 0.5):
    llen = len(lst)
    idx = int(pos * (llen - 1))
    vals = lst[idx:idx + 2]
    if len(vals) == 1:
        return vals[0]

    val1, val2 = vals
    ratio = (pos - idx / (llen - 1)) * (llen - 1)

    return val1 + (val2 - val1) * ratio


def luminance_mapper(fn: Callable, lumin_map: List[float]):

    @wraps(fn)
    def _lmapper(
        colors: List[Color], pos: float = 0.5, cspace: str = "lab", gamma: float = 1.0
    ) -> Color:
        lumin = linear_ip_f(lumin_map, pos)

        color = fn(colors, pos, cspace, gamma)
        return color.set(cie_l=lumin)

    return _lmapper


def project_domain(pos: float, domain: List[float]) -> float:
    """Project position of `pos` in `domain` into [0, 1] range
    """

    reverse = False
    if domain[0] > domain[-1]:
        # reverse
        domain = list(reversed(domain))
        reverse = True

    if pos <= domain[0]:
        return 0 if not reverse else 1
    if pos >= domain[-1]:
        return 1 if not reverse else 0

    lend = len(domain)
    prev = (0.0, 0.0)
    for posx, domx in enumerate(domain):
        curr = (posx / (lend - 1), domx)
        if pos <= domx:
            break
        prev = curr

    value = (pos - prev[1]) / (curr[1] - prev[1]) * (curr[0] - prev[0]) + prev[0]

    return value if not reverse else 1 - value


class ColorScale:
    """Maps numeric values to a color palette
    """

    DISPLAY_BORDER = 1
    DISPLAY_HEIGHT = 4
    DISPLAY_WIDTH = 40
    DISPLAY_STEPS = 12
    INTERPOLATORS = {
        "linear": linear_ip,
        "bezier": bezier_ip,
    }

    def __init__(
        self,
        colors: List[Any] = [Color("#ffffff"), Color("#000000")],
        domain: List[float] = None,
        gamma: float = 1.0,
        cspace: str = "lab",
        gamma_correction: float = None,
        interpolator: str = "linear",
        luminance_map: List[float] = None,
        cyclic: bool = False,
        name: str = None
    ):

        # convert 'colors' to list of Colors
        self.colors = [c if isinstance(c, Color) else Color(c) for c in colors]
        if cyclic:
            self.colors = self.colors + self.colors[0:1]
        self.domain: List[float] = [0, 1] if domain is None else domain
        self.gamma = gamma
        self.cspace = cspace
        self._luminance_map = luminance_map
        self._interpolator = self.INTERPOLATORS.get(interpolator, linear_ip)
        self._lumin_interpolator = None

        if not name:
            name = (self.colors[0].name + "_" + self.colors[-1].name).replace("#", "")

        self.name = name

        if gamma_correction is None:
            if (
                cspace in ["rgb", "hsl", "hsv", "hwb"]
                and self.interpolator != bezier_ip
            ):
                gamma_correction = 2.2
            else:
                gamma_correction = 1.0

        self.gamma_correction = gamma_correction

    @property
    def interpolator(self):
        if not self.luminance_map:
            return self._interpolator
        else:
            if not self._lumin_interpolator:
                self._lumin_interpolator = luminance_mapper(self._interpolator, self.luminance_map)
            return self._lumin_interpolator

    @interpolator.setter
    def interpolator(self, fn: Callable):
        self._interpolator = fn
        self._lumin_interpolator = None

    @property
    def luminance_map(self):
        return self._luminance_map

    @luminance_map.setter
    def luminance_map(self, lmap: List[float]):
        self._luminance_map = lmap
        self._lumin_interpolator = None

    @property
    def reversed(self):
        return self.domain[0] > self.domain[-1]

    def reverse(self):
        self.domain.reverse()

    def to_cmap(self, size: int = 256):
        """convert to matplotlib Colormap
        """
        from matplotlib.colors import ListedColormap  # type: ignore
        return ListedColormap([c.pltc for c in self.samples(size)], self.name, size)

    def __getitem__(self, key):
        if isinstance(key, (float, int)):
            return self._get_color_for_pos(key)
        # TODO slices

    def __str__(self):
        return f"[{self.name}]"

    def __repr__(self):
        return f"<ColorScale {self}>"

    def _get_color_for_pos(self, pos: float) -> Color:
        projpos = project_domain(pos, self.domain)

        if self.gamma != 1.0:
            projpos = projpos ** self.gamma

        colors = self.colors

        if projpos < 0:
            return colors[0] if not self.reversed else colors[-1]
        if projpos > 1:
            return colors[-1] if not self.reversed else colors[-1]

        interp = self.interpolator

        return Color(
            interp(self.colors, projpos, self.cspace, self.gamma_correction)
        )

    def samples(self, n: int = 10):
        return [self[self.domain[0] + (self.domain[-1] - self.domain[0]) * i / (n - 1)] for i in range(n)]

    def _displayimage(
        self,
        width: int = None,
        height: int = None,
        border: int = None,
        bgcolors: List["Color"] = None,
        steps: int = None,
    ) -> List[List["Color"]]:
        if width is None:
            width = self.DISPLAY_WIDTH
        if height is None:
            height = self.DISPLAY_HEIGHT
        if border is None:
            border = self.DISPLAY_BORDER
        if steps is None:
            steps = width

        w = width + 2 * border
        h = height + 2 * border
        img = []

        if bgcolors is None:
            bgcolors = getattr(
                self, "bgcolors", [Color(LabTuple(95, 0, 0)), Color(LabTuple(65, 0, 0))]
            )

        bgl = len(bgcolors)
        frm, to = (self.domain[0], self.domain[-1]) if not self.reversed else (self.domain[-1], self.domain[0])
        frm, to = frm * steps, to * steps

        for y in range(h):
            line = []
            for x in range(w):
                bgc = bgcolors[(y + x) % bgl]
                if x < border or y < border or x > w - border - 1 or y > h - border - 1:
                    line.append(bgc)
                else:
                    pos = math.floor(frm + (to - frm) * (x - border) / width) / (steps - 1)
                    line.append(
                        blend(
                            self[pos],
                            bgc,
                        )
                    )

            img.append(line)

        return img

    @property
    def N(self):
        return self.domain[-1] if not self.reversed else self.domain[0]

    @property
    def displayimage(self) -> List[List[Color]]:
        return self._displayimage()

    @property
    def termimage(self) -> str:
        return terminal.draw(self.displayimage)

    def print(
        self,
        fmt: str = "display",
        width: int = None,
        height: int = None,
        border: int = None,
        bgcolors: List["Color"] = None,
        force_ansi: bool = False,
        steps: int = None,
    ):
        if not force_ansi and not sys.stdout.isatty() and fmt == "display":
            fmt = "lhex"

        if fmt == "display":
            if border is None:
                border = self.DISPLAY_BORDER
            if width is None:
                width = terminal.termsize()[0] - 2 * border

            print(terminal.draw(self._displayimage(width, height, border, bgcolors, steps)))

        else:
            if steps is None:
                steps = self.DISPLAY_STEPS
            if steps < 2:
                steps = 2
            for i in range(steps):
                self[i/(steps - 1)].print(fmt=fmt)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['repacolors', 'repacolors.command', 'repacolors.palette', 'repacolors.themes']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0']

extras_require = \
{'xextras': ['xcffib>=0.9.0,<0.10.0']}

entry_points = \
{'console_scripts': ['repacolor = repacolors.command.repacolor:color']}

setup_kwargs = {
    'name': 'repacolors',
    'version': '0.8.0',
    'description': 'Small library for color conversion, manipulation, etc.',
    'long_description': '# repacolors\n\nSmall library for color conversion, manipulation, etc.\n\n[![Build Status](https://travis-ci.com/dyuri/repacolors.svg?branch=master)](https://travis-ci.com/dyuri/repacolors)\n\n![demo](./demo.svg)\n\n## Install\n\n```shell\n$ pip install repacolors\n```\n\nTo get the colors from `Xrdb`, install it with the `xextras` extras:\n\n```shell\n$ pip install repacolors[xextras]\n```\n\n## `repacolor` command\n\n```shell\n$ repacolor --help\nUsage: repacolor [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  adjust-contrast  Adjust the colors to match required contrast ratio.\n  colorwheel       Display colorwheel defined by `name` or created by the...\n  display          Display information about the provided colors.\n  palette          Get colors of given palette\n  pick             Pick colors from your desktop.\n  scale            Display color scale defined by the colors provided.\n```\n\n### `display`\n\nDisplay color information in the terminal.\n\n```shell\n$ repacolor display red\n\n+--------+ red - #ff0000\n|  BIG   | rgb(255, 0, 0)\n|  RED   | hsl(0, 100%, 50%)\n| SQUARE | lab(53.24% 80.09 67.2)\n+--------+\n\n$ repacolor display "#ffaad5" "rgb(128, 12, 46, .8)"\n... (displays both colors)\n$ echo "#ffffff" | repacolor display\n... (displays `white`)\n```\n\n### `pick`\n\nExecutes color picker and displays the picked color.\n\n```shell\n$ repacolor pick\n```\n\nThe integrated color picker works under _X11/linux_ if installed with *xextras*. If you want to use an external color picker, set the `COLORPICKER` environment variable:\n\n```shell\n$ export COLORPICKER=xcolor\n$ repacolor pick\n```\n\n### `palette`\n\nDisplay the colors of the palette. If no palette name provided, it shows the palettes available.\n\n```shell\n$ repacolor palette\nList of available palette names:\nryb, rybw3, orrd, pubu, ...\n\n$ repacolor palette viridis\n#440154\n#482777\n...\n```\n\n### `scale`\n\nDisplay a color scale defined by the input colors.\n\n```shell\n$ repacolor scale red white\n[colors from red to white]\n$ repacolor palette viridis | repacolor scale\n[color scale defined by `viridis` colors]\n```\n\n### `colorwheel`\n\nDisplay a color wheel.\n\nPre defined color wheels:\n\n- `ryb` - The RYB color wheel\n- `rgb` or `hsl` - The RGB color wheel\n- `lab` or `lch` - CIELAB color wheel\n\nIf no color wheel name provided, it will create one from the colors provided on `stdin`.\n\n```shell\n$ repacolor colorwheel rgb\n[RGB color wheel]\n$ repacolor scale red white black red | repacolor colorwheel\n[red - white - black color wheel]\n```\n\n### `adjust-contrast`\n\nAdjust the colors to match required contrast ratio.\n\nIf only one color is provided, chooses black or white based on the colors luminance.\n\nIf two colors are provided, tries to lighten/darken them to fulfill the contrast requirement. Starts to adjust the first color, then if it\'s not enough continues with the other.\n\nDefault contrast is 4.5 (WCAG AA). [More info on MDN.](https://developer.mozilla.org/en-US/docs/Web/Accessibility/Understanding_WCAG/Perceivable/Color_contrast)\n\n```shell\n$ repacolor adjust-contrast red --format=lhex\n#ff0000\n#000000  # chooses black for red\n$ repacolor adjust-contrast "#555" "#5e8d87" --format=lhex\n#1d1d1d  # #555 adjusted to be darker\n#5e8d87\n$ repacolor adjust-contrast "#5e8d87" "#555" --format=lhex\n#b6cfcc  # #5e8d87 lightened\n#555555\n$ repacolor adjust-contrast "#5e8d87" "#555" -v\nColors adjusted. (2.0007 => 4.5224)\n  #5e8d87   =>   #b6cfcc\n  #555555   =>   #555555\n```\n',
    'author': 'Gyuri Horak',
    'author_email': 'dyuri@horak.hu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dyuri/repacolors',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
